# Onboard classification 
 
The Onboard classification is made for loading a captured Hyperspectral Image (HSI) stored on the HYPSO satellite and generate a labeled image from it. 

It uses the machine learning algorithm Support Vector Machine (SVM). SVM are doing binary classifiaction. By doing multi-class classification Binary Desicion Tree is being used.  

For a captured HSI with 956x684 pixels will a labeled 
image have a size of following:

- 80kB with up to 2 classes.
- 160kB with up to 4 classes.
- 320kB with up to 16 classes.

The onboard classification is divided into three parts 

- Ground Training 
- Onboard Classification 
- Decoding 

</br>

# Ground Training 

This section will exlain how the machine learning algorithm SVMBDT
train with the labeled HSI. 

## Training Data

The training data that has been used to train the models, can be found at [url]. 

### Binary Descion Tree (BDT) Structures

Different BDT Structure has been designed to test for different scenarios.

The following structures is designed as followed:  

</br>

BDT 1

</br>

![bdt1](./figures/on_board_bdt1_.svg)



BDT 2



![bdt2](./figures/on_board_bdt2.svg)

</br>

BDT 3

</br>

![bdt3](./figures/on_board_bdt3.svg)

</br>

The SVM Models are trained with the function from ``svmbdt.py`` and 
the weights are produces by the jupyter notebooks: 

- ``svmbdt_training_bdt1.ipynb``
- ``svmbdt_training_bdt2.ipynb``
- ``svmbdt_training_bdt3.ipynb``

The training models with the weigths are stored in the ``svm_bdt/config/bdt/``

## Sparse SVM 

The bands selection for are stored in the ``svm_bdt/config/sparse_bands/``


</br>

# Onboard Implementation 

![Onboard classification figure](./figures/onboard_classification_pipeline_dark.svg)

The execution of the Onboard Classification: 

```c 

./onboard_classification.exe <capture folder file path> <output file path> <config folder prefix> <bdt index> <processing mode> <local angels path>

```

``bdt index `` is choosing the BDT design explained above. 

The numbers available are: 

| Number | BDT   |
|--------------|---------------|
| 1            | BDT 1 |
| 2            | BDT 2  |
| 3            | BDT 3  |

</br>

The argument ``processing mode`` has 4 different settings. 

These are: 

| Number | Processing mode    |
|--------------|---------------|
| 0            | Radiance         |
| 1            | Radiance Sparse |
| 2            | Reflectance  |
| 3            | Reflectance Sparse  |

</br>

The ``local angels path`` argument are just necessary for mode number ``2`` and ``3``. 
The local angels needs to be uplinked to the satellite and uses the solar zenith angle. 


## Radiometric calibration 

The radiometric calibration is meant for calibrating the
HSI Cube so that the ligth intencity from the same are the 
same for all of the images. 
The radiometric calibration of the cube is nessasary when 
classifiying since it will give surfaces similiar brigthness 
in the cubes. 

The public functions for the radiometric calibration implementation is ``radiometric_calibration_init()`` and ``radiometric_calibration_pixel()``. 

- ``radiometric_calibration_init()`` loads the nessecary data to the structs and metadata to usage for calibration. 
- ``radiometric_calibration_pixel()`` calibrates one pixel (120 spectral bands). (Calibrates pixelwise due to using much less memory, instead of allocating memory to the whole HSI Cube). 

```c
int8_t radiometric_calibration_init(
    char capture_path[MAX_CHAR_PATH], 
    char config_folder_path[MAX_CHAR_PATH], 
    char local_angels_path[MAX_CHAR_PATH],
    uint16_t processing_mode, 
    float radiometric_metadata[3], 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    struct extraterrestrial_solar_irradiance *esi, 
    struct local_angels *la);

int8_t radiometric_calibration_pixel(
    struct hsi_cube_pixel *cube_pixel, 
    struct calibrated_cube_pixel *calibrated_pixel, 
    uint16_t processing_mode, 
    float exposure, 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    struct extraterrestrial_solar_irradiance *esi, 
    struct local_angels *la,
    float background_value, 
    float earth_sun_distance, 
    uint32_t radio_index, 
    uint16_t local_angels_index);
```

## Support Vector Machine with Binary Desicion Tree (SVMBDT)

The SVMBDT is machine learning algorithm that are using linear seperation 
in a three structure (Described above) to determine a class (Classes defined below: Decode). 
It is doing it pixelwise making it possible to have different ratios and dimension when capture a HSI. 

The public function for the SVMBDT is 
``svmbdt_init()`` and ``svmbdt_predict()``. 

- ``svmbdt_init()`` loads the bdt, svm model and sparse data necessary for doing the classification of the HSI. 
- ``svmbdt_predict()`` predicts one pixel and writes the label to the ``struct Labeled_Image* image``. 

```c
int8_t svmbdt_init(
    char config_folder_path[MAX_CHAR_PATH_LENGTH], 
    uint8_t bdt_num, 
    uint8_t processing_mode, 
    uint8_t bdt_model[MAX_CLASSES][BDT_INDEX_LENGTH], 
    float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
    uint8_t svmbdt_metadata[1]);


int8_t svmbdt_predict(
    float X[SPECTRAL_BAND_SIZE], 
    uint8_t bdt[MAX_CLASSES][BDT_INDEX_LENGTH], 
    struct Labeled_Image* image, 
    float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
    uint8_t n_bands_used, 
    uint8_t processing_mode, 
    uint32_t index);


int8_t labeled_image_bin(char labels_save_path[MAX_CHAR_PATH_LENGTH],
    struct Labeled_Image *image,
    uint32_t n_pixels);
```

## Config Folder

The config folder are placed in ``onboard_classification/config/``

The folder folder strucutre is shown below: 

```
├── bdt
    ├── bdt1
    ├── bdt2
    ├── bdt3
├── calibration_files    
    ├── rad_coeffs_FM_binx9.csv
    ├── spectral_coeffs_FM_order2.csv
├── constants
    ├── mean_extraterrestrial_solar_irradiance_chkur
    ├── mean_extraterrestrial_solar_irradiance_thuillier
├── sparse_bands
    ├── sparse_bands1
    ├── sparse_bands2
    ├── sparse_bands3
├── svm_models
    ├── svm_models1
        ├── lsm_radiance
           ├── lsm0001
           ...
           └── lsm0809
        ├── lsm_radiance_sparse
        ├── lsm_reflectence
        └── lsm_reflectence_sparse
    ├── svm_models2
    ├── svm_models3
```

``lsm`` in the ``svm_models/`` folder stands for linear svm model. These contains the two classes, the intercept and the weigths. The file structure is: 

``[class 1 (uint8_t), class 2 (uint8_t), intercept (Float), weigths (120)*(Float)]``. 

</br>

# Decoding 

The labeled image are stored as a binary file. 
The python program ``decode_labels.py`` will decode the binary file 
and store .png and .npy files of the labeled image.

The python script is executed as followed with default dimensions: 

The default dimensions are: 

| x dim | y dim    | 
|--------------|---------------|
| 684            | 956         | 
| 598            | 1092 | 
| 537            | 1216  | 



```console
python3 decode_labels.py <compressed labels bin file path> <decoded labels save path>
```
With custom dimension, print metadata and skew mode: 

```console
python3 decode_labels.py <compressed labels bin file path> <decoded labels save path> <x dim> <y dim> <print metadata> <skew mode>
```

For one of the default HSI dimension: ``x_dim`` = 684 and  ``y_dim`` = 956. 

``print_execution_time`` : 0, just creates the decode labels. 

``print_execution_time`` : 1, Prints the loading and classification execution time stored in the labels packet in console. 

``skew`` : 0, full dimensions of the labeled image. 

``skew`` : 1, format the labeled image to bin3 format. Example 956x684 -> 956x228.  

When using ``skew``, ``y_dim`` and ``x_dim`` are still 956 and 684 when writing in console.  

## Classes

The labeles in the image are numbered with
different classes with RGB values for plotting as shown below: 

| Class Number | Class Name    | R |G | B |
|--------------|---------------|--------------|---------------|---------------|
| 0            | Water         | 0 | 0 | 139|
| 1            | Strange Water | 13| 152 | 186 |
| 2            | Light Forest  | 144 | 238 | 144 |
| 3            | Dark Forest   | 20| 70 | 50 |
| 4            | Urban         | 240 | 0 | 0 |
| 5            | Rock          | 107 | 74 | 58 |
| 6            | Ice           | 255 | 255 | 255 | 
| 7            | Sand          | 194 | 178 | 128 | 
| 8            | Thick Clouds  | 180 | 180 | 180 | 
| 9            | Thin Clouds   | 100 | 100 | 100 | 
| 10           | Shadow        | 0 | 0 | 0 | 

</br>

The color of the corresponding class on the labels:

<img src=figures/hypso_legend.png alt="Hypso Legend" width="230"
    height="350" title="Hypso Title">


## File Structure

The labeled image binary file packet structure is shown below: 

![Labels File Structure](./figures/labels_file_structure_dark.svg)