"""
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
"""

import numpy as np
import copy
from HelperFunctions_ import * 
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
import struct
import sklearn.metrics

"""
This file contains the functions used for 
groundtraining for the svmbdt onboard model
and also the analysing of the labeled binary files 
generated by the 

# 0 Water 
# 1 Strange Water 
# 3 Forest 
# 4 Urban 
# 5 Rock 
# 6 Ice 
# 7 Sand
# 8 Thick Clouds 
# 9 Thin Clouds 
# 10 Shadow

"""


def index_svm_model(next_model, tree_branch, svm_model_place): 
    # Find the index for next tb, if not exisiting return -1 
    # Find the index where the next svm_model is placed. 
    svm_model_index = 0
    svm_model_name_tb_index = 0 

    for i in range(len(tree_branch)):
        if(tree_branch[i][0] == next_model): 
            svm_model_name_tb_index = i 
            break

    for i in range(len(svm_model_place)): 
        if(svm_model_place[i] == next_model):
            svm_model_index = i
            break

    return svm_model_index, svm_model_name_tb_index


def float_to_int_delta(value, byte):
    
    import copy

    max_value = 50
    v = copy.deepcopy(value)
    delta = max_value/(pow(2,8*byte)-1)
    converted_int = int(v/delta)

    if(converted_int >= (pow(2,8*byte)/2)-1): 
        return int(pow(2,8*byte)/2)-1
    
    elif(converted_int <= -(pow(2,8*byte)/2)+1): 
        return -int(pow(2,8*byte)/2)+1
    
    else: 
        return converted_int

def int_to_float_delta(value, byte): 
    import copy
    max_value = 50
    v = copy.deepcopy(value)
    delta = max_value/(pow(2,8*byte)-1)
    return v*delta

def int_to_svm_model_name(value): 
    first_value = int((value - value%16)/16)
    second_value = value%16
    return str(first_value), str(second_value) 

def string_int_to_hex_int(str_hex):
    
    if(len(str_hex) == 2): 
        return 16*int(str_hex[0]) + int(str_hex[1])
    
    if(len(str_hex) == 4): 
        return 16*int(str_hex[0:2]) + int(str_hex[2:4])


def linear_train_int_bin(file_path, class_1, class_2, b, w): 
        #svm_model format: [class1, class2, b, w_i[]]

    #Modus: type of svm model. Linear, rbf, sparse etc.

    import struct 
    
    svm_model_name = ""

    c1 = str(class_1) 
    c2 = str(class_2)

    if(len(c1) == 1): 
        c1 = "0" + c1

    if(len(c2) == 1): 
        c2 = "0" + c2

    svm_model_name = "lsm" + c1 + c2

    file = open(file_path + svm_model_name, "wb")    
    
    buf = struct.pack("BB", class_1, class_2)

    file.write(buf)

    buf = struct.pack('h', b)

    file.write(buf)

    w_size = str(len(w)) + "h"

    buf = struct.pack(w_size, *w)
    
    file.write(buf)

    file.close()

    return 0


def linear_train_to_bin(file_path, class_1, class_2, b, w): 
        #svm_model format: [class1, class2, b, w_i[]]

    #Modus: type of svm model. Linear, rbf, sparse etc.

    import struct 
    
    svm_model_name = ""

    c1 = str(class_1) 
    c2 = str(class_2)

    if(len(c1) == 1): 
        c1 = "0" + c1

    if(len(c2) == 1): 
        c2 = "0" + c2

    svm_model_name = "lsm" + c1 + c2

    file = open(file_path + svm_model_name, "wb")    
    
    buf = struct.pack("BB", class_1, class_2)

    file.write(buf)

    buf = struct.pack('f', b)

    file.write(buf)

    w_size = str(len(w)) + "f"

    buf = struct.pack(w_size, *w)
    
    file.write(buf)

    file.close()

    return 0


def bdt_config_bin(file_path, bdt_config_aray, config_num):
    
    import struct 

    bca = np.array(bdt_config_aray).flatten()

    bdt_config_name = "bdtc" + str(config_num)
    
    file = open(file_path + bdt_config_name, "wb")

    for i in range(len(bca)):
        buf = struct.pack(">B", bca[i]) 
        file.write(buf)

    file.close()


def hsi_binary(file_path, filename, X):
    
    import struct

    pH = np.array(X).flatten()
        
    file = open(file_path + filename, "wb")

    for i in range(len(pH)):
        buf = struct.pack("h", float_to_int_delta(pH[i],2)) 
        file.write(buf)

    file.close()

def models_names(bdt, model_names): 
    
    svm_model_list = []

    for i in range(len(bdt)): 
        c1, c2 = int_to_svm_model_name(bdt[i][0])
        
        if(len(c1) == 1): 
            c1 = "0" + c1

        if(len(c2) == 1): 
            c2 = "0" + c2

        svm_model_list.append(str(model_names) + str(c1) + str(c2)) 

    return svm_model_list


def rotate_image(image): 
    rotated_image = []
    for i in range(len(image[0])): 
        x_array = []
        for j in range(len(image)-1,-1,-1):
            x_array.append(image[j][i])
        rotated_image.append(x_array)

    return np.array(rotated_image)


def svm_branch_models_to_bin(svm_branch_model, b_extra, path):
    
    sbm = copy.deepcopy(svm_branch_model)
    class_1 = sbm.classes_[0]
    class_2 = sbm.classes_[1]
    b = float_to_int_delta(sbm.intercept_[0] + b_extra ,2)
    w = sbm.coef_[0]
    tw = []

    for i in range(len(w)): 
        tw.append(float_to_int_delta(w[i],2))

    linear_train_int_bin(path, class_1, class_2, b, tw)


def bdt_string_to_config_format(bdt):
    
    import copy

    config_bdt = []

    for i in range(len(bdt)): 
        bdt_index = []
        bdt_i = copy.deepcopy(bdt[i][0]) 

        bdt_index.append(string_int_to_hex_int(bdt_i))
        bdt_index.append(int(bdt[i][1]))
        bdt_index.append(int(bdt[i][2]))
        config_bdt.append(bdt_index)

    return config_bdt


def store_labeled_image(predicted_label, x_dim, y_dim, path): 

    hypso_color_label = { 0 : [0,0,139], 
                            1 : [13,152,186], 
                            2 : [144,238,144], 
                            3 : [20,70,50], 
                            4 : [240,0,0], 
                            5 : [107, 74, 58], 
                            6 : [255,255,255],  
                            7 : [194,178,128], 
                            8 : [100,100,100], 
                            9 : [180,180,180],  
                            10 : [0,0,0]
    }


    new_predicted_label = copy.deepcopy(predicted_label)
    new_predicted_label = np.array(new_predicted_label)

    np.save(path, new_predicted_label)

    yout_color = newColorLabels(new_predicted_label,x_dim,y_dim, hypso_color_label)

    plt.imshow(yout_color)
    plt.axis('off')
    plt.savefig(path, dpi = 300, bbox_inches="tight", pad_inches = 0)
    plt.close()
        

def fixing_labels(pl, x_0, x_1, y_0, y_1, x_dim, y_dim, old_label, new_label): 
    fixel_label = np.array(pl)
    fixel_label.resize(x_dim, y_dim)

    for i in range(len(fixel_label)): 
        for j in range(len(fixel_label[0])): 
            if((i >= x_0) and i <= x_1): 
                if(j >= y_0 and j <= y_1): 
                    if(fixel_label[i][j] == old_label): 
                        fixel_label[i][j] = new_label

    return fixel_label.flatten()

  

def rotate_labels(labeled_image, x_dim, y_dim):
    
    resized_li = np.array(labeled_image)
    resized_li.resize(y_dim,x_dim)
    rotate_image = []

    for i in range(len(resized_li[0])): 
        for j in range(len(resized_li)-1,-1,-1): 
            rotate_image.append(resized_li[j][i])
    return np.array(rotate_image) 

def rotate_labels2(labeled_image, x_dim, y_dim):
    
    resized_li = np.array(labeled_image)
    resized_li.resize(y_dim,x_dim)
    rotate_image = []

    for i in range(len(resized_li[0])): 
        for j in range(len(resized_li)): 
            rotate_image.append(resized_li[j][i])
    return np.array(rotate_image) 

def rotate_labels3(labeled_image, x_dim, y_dim):
    
    resized_li = np.array(labeled_image)
    resized_li.resize(y_dim,x_dim)
    rotate_image = []

    for i in range(len(resized_li)-1,-1,-1): 
        for j in range(len(resized_li[0])): 
            rotate_image.append(resized_li[i][j])
    return np.array(rotate_image) 

def minimize_pixels(data): 
    new_data = []
    for i in range(len(data)): 
        x_list = []
        for j in range(0,len(data[0]),3):
            x_list.append(data[i][j])
        new_data.append(x_list)
    return new_data

def processed_hsi(hsi_cube_path,hsi_labels_path): #For HYSPO data with 120 spectral bands
    data = np.load(hsi_cube_path)
    new_data = minimize_pixels(data)
    X = np.array(new_data).reshape((-1,120))

    labels = np.load(hsi_labels_path)
    new_labels = rotate_labels(labels, len(new_data), len(new_data[0]))
    #new_labels = labels
    y = np.array(new_labels).flatten()

    return X,y

def processed_hsi_data(hsi_cube_path): #For HYSPO data with 120 spectral bands
    data = np.load(hsi_cube_path)
    new_data = minimize_pixels(data)
    X = np.array(new_data).reshape((-1,120))
    return X


def bdt_to_bin(path, data): 

    data_flatten = np.array(data).flatten()
        
    file = open(path, "wb")

    buf = struct.pack(str(len(data_flatten)) + 'B',*(data_flatten))
    file.write(buf)

    file.close()

def sparse_bands_to_bin(path, data): 
        
    file = open(path, "wb")

    buf = struct.pack(str(len(data)) + 'B',*(data))
    file.write(buf)

    file.close()


def svm_branch_models_to_bin(svm_branch_model, b_extra, path):
    
    sbm = copy.deepcopy(svm_branch_model)
    class_1 = sbm.classes_[0]
    class_2 = sbm.classes_[1]
    b = float_to_int_delta(sbm.intercept_[0] + b_extra ,2)
    w = sbm.coef_[0]
    tw = []

    for i in range(len(w)): 
        tw.append(float_to_int_delta(w[i],2))

    linear_train_int_bin(path, class_1, class_2, b, tw)


def read_compressed_labels(file_path):
     
     #Functions that analyse the binary file generated by the SVMBDT 
     #and converts it to a readable execution [int], n_classes[], and 
     #the predicted labels[].
     
    with open(file_path, mode='rb') as file: # b is important -> binary
        fileContent = file.read()
    filesize = len(fileContent)
    unpack_opt = "ll16B" + str(filesize-4-4-16) + "B"
    arr = struct.unpack(unpack_opt, fileContent)

    classification_execution_time = arr[0]
    radiometric_execution_time = arr[1]
    classes_holder = arr[2:18]
    labels_holder = arr[18:]

    classes = []
    labels = []

    # Decode the labels and converts 
    # them back to original classes. 

    for i in range(len(classes_holder)): 
        if(int(classes_holder[i]) != 255): 
            classes.append(classes_holder[i])

    if(len(classes) <= 2): 
        for i in range(len(labels_holder)): 
            pixel_str = format(labels_holder[i], "b").zfill(8)
            for j in range(8): 
                labels.append(int(pixel_str[j],2))

    if(len(classes) > 2 and len(classes) <= 4): 
        for i in range(len(labels_holder)): 
            pixel_str = format(labels_holder[i], "b").zfill(8)
            for j in range(4): 
                labels.append(int(pixel_str[2*j:(j+1)*2],2))

    if(len(classes) > 4 and len(classes) <= 16):
        for i in range(len(labels_holder)): 
            pixel_str = format(labels_holder[i], "b").zfill(8)
            for j in range(2): 
                labels.append(int(pixel_str[4*j:(j+1)*4],2)) 

    for i in range(len(labels)): 
        labels[i] = classes[labels[i]]

    return classification_execution_time, radiometric_execution_time, classes, labels


def bin_results_to_numpy_and_png(results_path, save_path): 

    from os import listdir
    from os.path import isfile, join
    files = [f for f in listdir(results_path) if isfile(join(results_path, f))]

    for file in files: 
        execution_time, classes, labels = read_compressed_labels(results_path + file)
        store_labeled_image(labels,956,228*3, save_path + file)

def combineDataAndLabels(data_array, labels_array):
    combined_data = []
    combined_labels = []

    if(len(data_array) == len(labels_array)):
        for i in range(len(data_array)):
            for j in range(len(data_array[i])): 
                combined_data.append(data_array[i][j])
                combined_labels.append(labels_array[i][j])
    
    return np.array(combined_data), np.array(combined_labels) 


def array_to_bin(path, data): 

    data_flatten = np.array(data).flatten()
        
    file = open(path, "wb")

    buf = struct.pack(str(len(data_flatten)) + 'f',*(data_flatten))
    file.write(buf)

    file.close()

def reduce_and_rotate_pixels(cube, rotate=0): 
    new_cube = np.zeros((684,956,120))

    if(rotate == 0):
        for i in range(684):
            new_cube[i,:,:] = cube[:,i,:]

    if(rotate == 1): 
        for i in range(684):
            new_cube[i,:,:] = cube[:,683-i,:]

    minimized_cube = np.zeros((228,956,120))
    index = 0
    for i in range(0,684,3): 
        minimized_cube[index,:,:] = new_cube[i,:,:]
        index+=1

    return minimized_cube

def plot_cube(cube, path, x_dim, y_dim):
    new_cube = np.zeros((x_dim,y_dim,3))
    new_cube[:,:,0] = cube[:,:,32]
    new_cube[:,:,1] = cube[:,:,45]
    new_cube[:,:,2] = cube[:,:,56]
    new_cube /= new_cube.max()/2
    new_cube[new_cube > 1 ] = 1

    plt.imshow(new_cube)
    plt.axis('off')
    plt.savefig(path, dpi = 300, bbox_inches="tight")
    plt.close()

def process_hsi_to_training(cube_path, plot_path, save_path, rotate=0): 
    cube = np.load(cube_path)
    new_cube = reduce_and_rotate_pixels(cube, rotate=rotate)
    plot_cube(new_cube, plot_path)
    X = np.array(new_cube).reshape(-1,120)
    array_to_bin(save_path, X)

    return X

def bin_results_to_numpy_and_png(results_path, save_path): 

    from os import listdir
    from os.path import isfile, join
    files = [f for f in listdir(results_path) if isfile(join(results_path, f))]

    for file in files: 
        execution_time, classes, labels = read_compressed_labels(results_path + file)
        store_labeled_image(labels,228,956, save_path + file)

def combineDataAndLabels(data_array, labels_array):
    combined_data = []
    combined_labels = []

    if(len(data_array) == len(labels_array)):
        for i in range(len(data_array)):
            for j in range(len(data_array[i])): 
                combined_data.append(data_array[i][j])
                combined_labels.append(labels_array[i][j])
    
    return np.array(combined_data), np.array(combined_labels) 


def mean_extraterrestrial_solar_irradiance_to_bin(file_path, arr): 

    import struct 

    file = open(file_path, "wb") 
    
    arr_size = str(len(arr)) + "f"
    
    buf = struct.pack(arr_size, *arr)
    
    file.write(buf)

    file.close()

    return 0

def load_satellite_cube_bin(path): 

    # Are used to load the radiometric calibrated cubes. 
    # Just for testing purposes. 
    # NOT for opening bip file cubes. 
    
    import struct 
    import os

    n_spectral_bands = 120

    file = open(path, "rb") 
    
    file_size = str(int(os.stat(path).st_size/4)) + "f" 

    content = struct.unpack(file_size, file.read())

    cube = []

    for i in range(int(len(content)/n_spectral_bands)): 
        cube.append(content[i*n_spectral_bands:(i+1)*n_spectral_bands])
    
    file.close()

    return cube

def check_cube_difference(cube_1, cube_2):
    #For testing the difference between the two algorithms from 
    #python script vs the satellite code script.


    if(len(cube_1) == len(cube_2)): 
            print("Their are difference between the cubes!")

            cube_difference = np.abs(cube_1 - cube_2) 
            
            mean_difference = np.mean(cube_difference)
            max_difference = np.max(cube_difference)
            min_difference = np.min(cube_difference)


            print("Mean difference: ", round(mean_difference,6))
            print("Max difference : ", round(max_difference,6))
            print("Min difference : ", round(min_difference,6))
            print("Corrects: ", 
                round(100*(len(cube_difference[cube_difference <=  1e-1])
                           /len(cube_difference.flatten())),3)
                           ,"% under a threshold of 1e-1.")
        

            print("Corrects: ", 
                round(100*(len(cube_difference[cube_difference <=  1e-2])
                           /len(cube_difference.flatten())),3)
                           ,"% under a threshold of 1e-2.")
    
            print("Corrects: ", 
                round(100*(len(cube_difference[cube_difference <=  1e-3])
                           /len(cube_difference.flatten())),3)
                           ,"% under a threshold of 1e-3.")


   # return mean_difference, min_difference, max_difference


def plot_hypso_legend(path):
    import matplotlib.patches as mpatches
    import pylab

    hypso_color_label = { 0 : [0,0,139], 
                            1 : [13,152,186], 
                            #2 : [144,238,144], 
                            2 : [20,70,50], 
                            3 : [240,0,0], 
                            4 : [107, 74, 58], 
                            5 : [255,255,255],  
                            6 : [194,178,128], 
                            7 : [100,100,100], 
                            8 : [180,180,180],  
                            9 : [0,0,0]
    }

    hypso_classes = {0: "Water", 1: "Strange Water", #2: "Light Forest", 
                    2:"Forest", 
                    3 : "Urban", 4 : "Rock", 5 : "Ice/Snow", 6 : "Sand", 
                    7 : "Thick Clouds", 8 : "Thin Clouds", 9 : "Shadows"}
    
    afont = {'fontname':'Georgia'}

    cmap = {}

    for i in range(len(hypso_color_label)): 
        cmap[i] = np.array(hypso_color_label[i] + [255])/255

    patches = [mpatches.Patch(color=cmap[i], label=hypso_classes[i]) for i in hypso_color_label] 

    figLegend = pylab.figure(figsize = (1.65,2.25))
    figLegend.legend(handles=patches,loc = 'upper left')
    figLegend.savefig(path, dpi=1000)


def combineDataAndLabels(data_array, labels_array):
    combined_data = []
    combined_labels = []
    
    if(len(data_array) == len(labels_array)):
        for i in range(len(data_array)):
            for j in range(len(data_array[i])): 

                combined_data.append(data_array[i][j])
                combined_labels.append(labels_array[i][j])
    else: 
        print("The two arrays is not the same size")
    
    return np.array(combined_data), np.array(combined_labels) 

def plot_cube(cube, path, n_trainin_data = 8):
    new_cube = np.zeros((228*n_trainin_data,956,3))
    new_cube[:,:,0] = cube[:,:,70]
    new_cube[:,:,1] = cube[:,:,40]
    new_cube[:,:,2] = cube[:,:,20]
    new_cube /= new_cube.max()/2
    new_cube[new_cube > 1 ] = 1

    plt.imshow(new_cube)
    plt.axis('off')
    plt.savefig(path, dpi = 300, bbox_inches="tight", pad_inches = 0)
    plt.close()


def bin_results_to_numpy_and_png(results_path, save_path): 

    from os import listdir
    from os.path import isfile, join
    files = [f for f in listdir(results_path) if isfile(join(results_path, f))]

    for file in files: 
        execution_time, classes, labels = read_compressed_labels(results_path + file)
        if(int(len(labels)/956) > 228): 
            store_labeled_image(labels,956,int(len(labels)/956), save_path + file)

        else: 
            store_labeled_image(labels,int(len(labels)/956),956, save_path + file)


def training_data_to_binary_file(file_path, svm_branch_model, b_mode = 1): 

    svm_model_name = ""

    class_1_str = str(svm_branch_model.classes_[0])
    class_2_str = str(svm_branch_model.classes_[1])

    class_1 = svm_branch_model.classes_[0]
    class_2 = svm_branch_model.classes_[1]

    if(len(class_1_str) == 1): 
        class_1_str = "0" + class_1_str

    if(len(class_2_str) == 1): 
        class_2_str = "0" + class_2_str

    svm_model_name = "lsm" + class_1_str + class_2_str

    file = open(file_path + svm_model_name, "wb")    
    
    buf = struct.pack("BB", class_1, class_2)

    file.write(buf)

    if(b_mode == 1):

        b = svm_branch_model.intercept_[0]

        buf = struct.pack(f'{len([1])}f', *[b])

    else: 
        buf = struct.pack(f'{len([1])}f', *[0])

    w_i = svm_branch_model.coef_[0]

    file.write(buf)

    buf = struct.pack(f'{len(w_i)}f', *w_i)

    file.write(buf)

    file.close()

def process_hsi_to_training_no_plot(cube_path, rotate=0, bin_mode = 0):
    
    if(bin_mode == 1): 
        cube = np.array(load_satellite_cube_bin(cube_path))
        cube = cube.reshape(956,int(len(cube)/956),120)
    else:
        cube = np.load(cube_path)
    
    new_cube = reduce_and_rotate_pixels(cube, rotate=rotate)
    X = np.array(new_cube).reshape(-1,120)

    return X


def sparse_selection(combined_data, combined_labels): 

    import ssvm 

    X = copy.deepcopy(combined_data)
    y = copy.deepcopy(combined_labels)

    X = preprocessing.scale(X, axis=0)


    X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.9999, random_state=1)
    print(X_train.shape, X_test.shape, y_train.shape, y_test.shape)

    ssvm1 = ssvm.SSVM(eps=1, verbose=True, use_joint_matrix=False, decision_function='custom', Lambda=0.95,
                    groups=[[6,7,8,9],[0,1,3,4,5,10], [[7],[6,8,9]],[[0,1],[3,4,5,10]]])

    ssvm1.fit(X_train, y_train, do_continue=False)

    keys = list(ssvm1._w())

    keys = list(ssvm1._w())
    w1 = [ssvm1._w()[i] for i in keys]

    new_w1 = np.zeros((10,120))

    for i in range(len(w1)):
        new_w1[i][0] = 0 
        for j in range(1,len(w1[0])-1): 
            last_index = w1[i][j-1]
            if(np.abs(last_index - w1[i][j]) < 0.125): 
                new_w1[i][j] = 0
            else: 
                new_w1[i][j] = w1[i][j]

    add_new_w1 = []

    for i in range(len(new_w1[0])): 
        w1_index = 0
        for j in range(len(new_w1)):
            w1_index += new_w1[j][i]
        add_new_w1.append(w1_index)

    #plt.plot(add_new_w1)

    sparse_bands_index = []

    for i in range(len(add_new_w1)):
        if(add_new_w1[i] != 0): 
            sparse_bands_index.append(i)

    for i in range(120-len(sparse_bands_index)):
        sparse_bands_index.append(0)



    return sparse_bands_index, new_w1
    

def sparse_training_data(combined_data, sparse_band_index_array):
    
    sparse_combined_data = np.zeros((len(combined_data), len(combined_data[0]))) 
 
    for j in range(len(sparse_combined_data[0])): 
        if(sparse_band_index_array[j] != 0): 
            sparse_combined_data[:,sparse_band_index_array[j]] = combined_data[:,sparse_band_index_array[j]]

    return sparse_combined_data


def plot_hypso_confusion_matrix(gt_labels, predicted_label, fig_name = "confusion_matrix_test", bdt_mode = 1, f1_score = 0, multilabels = 0): 
    import seaborn as sns
    
    from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay, multilabel_confusion_matrix

    if(int(len(predicted_label)/956) == 684): 
        new_predicted_label = []
        for i in range(0,len(predicted_label), 3*956): 
            new_predicted_label.append(predicted_label[i:i+956])
        new_predicted_label = np.array(new_predicted_label).flatten()

    else: 
        new_predicted_label = predicted_label

    if((bdt_mode == 1) or (bdt_mode == 3)): 
        hypso_classes = ["Water", "Strange Water", "Light Forest", "Forest", 
                        "Urban", "Rock", "Ice/Snow", "Sand", 
                        "Thick Clouds", "Thin Clouds", "Shadows"]
        
    else: 
        hypso_classes = ["Water", "", "", "Land", 
                        "", "", "Ice/Snow", "", 
                        "Clouds", "", ""]
        
    if(multilabels == 1):
        cm = np.array(multilabel_confusion_matrix(gt_labels,new_predicted_label))

        classes = classesInLabels(gt_labels)

        if(len(classes) <= len(classesInLabels(new_predicted_label))): 
            classes = classesInLabels(new_predicted_label)

        n_classes = len(classes)
        
        if(n_classes >= 9): 
            figsize = (4.5,10)

        elif(n_classes >= 7): 
            figsize = (3.75,8)
        
        elif(n_classes >= 5):
            figsize = (3.5,6)
        
        if(n_classes%2==0): 
            if(n_classes < 3): 
                figsize = (2,5)
                fig, axs = plt.subplots(n_classes,1, figsize=figsize)

            elif(n_classes < 5): 
                figsize = (2,8)
                fig, axs = plt.subplots(n_classes,1, figsize=figsize)
            else: 
                fig, axs = plt.subplots(int(n_classes/2),2, figsize=figsize)
        else: 
            if(n_classes < 5): 
                figsize = (2.2,6)
                fig, axs = plt.subplots(n_classes,1, figsize=figsize)
            else: 
                fig, axs = plt.subplots(int(n_classes/2)+1,2, figsize=figsize)
        
        classes_index = 0

        for confusion_matrix, ax in zip(cm, axs.flatten()): 
            if(classes_index < n_classes):
                ConfusionMatrixDisplay(np.flip(confusion_matrix), 
                                    display_labels=["P","N"]).plot(ax=ax, colorbar=0, cmap="seismic")
                ax.title.set_text(hypso_classes[classes[classes_index]])
                classes_index += 1

        for ax in axs.flat: 
            if not bool(ax.has_data()): 
                fig.delaxes(ax)
            
        plt.tight_layout()
        plt.savefig(fig_name, dpi=300)
        plt.close()

    else: 

        classes = classesInLabels(list(classesInLabels(gt_labels)) + list(classesInLabels(new_predicted_label)))

        classes_names = []
        for i in classes: 
            classes_names.append(hypso_classes[i])
        
        afont = {'fontname':'Georgia'}

        cm = confusion_matrix(gt_labels,new_predicted_label) 

        fig = plt.figure(figsize=(8, 7))
        ax= plt.subplot()
        sns.heatmap(cm, annot=True, ax = ax, fmt = 'g', cmap="Blues")
        ax.set_xlabel('Predicted', fontsize=20, **afont)
        ax.xaxis.set_label_position('bottom')
        plt.xticks(rotation=90)
        ax.xaxis.set_ticklabels(classes_names, fontsize = 10, **afont)
        ax.xaxis.tick_bottom()

        ax.set_ylabel('True', fontsize=20, **afont)
        ax.yaxis.set_ticklabels(classes_names, fontsize = 10,  **afont)
        plt.yticks(rotation=0)

        plt.title('Confusion Matrix', fontsize=20, **afont)
        
        plt.tight_layout()
        plt.savefig(fig_name, dpi=300)
        plt.close()

    if(f1_score == 1): 
        #overall_accuracy = sklearn.metrics.f1_score(gt_labels, new_predicted_label, average="micro")
        overall_accuracy = sklearn.metrics.accuracy_score(gt_labels, new_predicted_label)

    else: 
        overall_accuracy = np.sum(gt_labels == new_predicted_label)/len(gt_labels)

    return overall_accuracy