/*
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
*/

#ifndef RADIOMETRIC_H
#define RADIOMETRIC_H

#define COMPILE_DO_DEBUG 1
 
#if COMPILE_DO_DEBUG
#define DEBUG(str) str
#else
#define DEBUG(str)
#endif // COMPILE_DO_DEBUG

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>

//Macro values
#define SPECTRAL_BAND_SIZE 120
#define MAX_HYPSO_CUBE_SIZE 956*684
#define MAX_CHAR_PATH 200
#define PI 3.141592653589793
#define SPECTRAL_CALIBRATION_LENGTH 5
#define DISTANCE_TABLE_SIZE 26
#define CUBE_Y_LENGTH 684
#define CUBE_X_LENGTH 956
#define DARK_FRAME_LENGTH 1080
#define CUBE_SIZE 956*684
#define FRAME_TIME_INDEX 1
#define SOLAR_ZENITH_ANGLE_INDEX 4
#define LATITUDE_INDEX 2
#define LONGITUDE_INDEX 3
#define DARK_FRAME_BUFFER_SIZE 36864
#define DARK_FRAME_BACKGROUND_VALUE 8.0

//Macro paths

#define SPECTRAL_CALIBRATION_PATH "/calibration_files/spectral_coeffs_FM_order2.csv"
#define RADIOMETRIC_COEF_PATH "/calibration_files/rad_coeffs_FM_binx9.csv"
#define DARK_PATH "/calibration_files/dark_frame_FM.csv"

#define IRRADIANCE_CHKUR_PATH "/constants/mean_extraterrestrial_solar_irradiance_chkur"
#define IRRADIANCE_THUILLIER_PATH "/constants/mean_extraterrestrial_solar_irradiance_thuillier"
#define CAPTURE_CONFIG_NAME "/capture_config.ini"


//Structs

struct hsi_cube_pixel{
    uint16_t X[SPECTRAL_BAND_SIZE]; 
};

struct calibrated_cube_pixel{
    float X[SPECTRAL_BAND_SIZE]; 
};

struct Spectral_calibration{
    float spectral_calibration[SPECTRAL_CALIBRATION_LENGTH];
};

struct radiometric_calibration_coeff{
    float coeff[CUBE_Y_LENGTH][SPECTRAL_BAND_SIZE];
};

struct dark_frame{
    float frame[CUBE_Y_LENGTH][DARK_FRAME_LENGTH];
};

struct extraterrestrial_solar_irradiance{
    float coeff[SPECTRAL_BAND_SIZE];
};

struct local_angels{
    float coeff[CUBE_X_LENGTH][8];
};


int8_t radiometric_calibration_init(
    char capture_path[MAX_CHAR_PATH], 
    char config_folder_path[MAX_CHAR_PATH], 
    char local_angels_path[MAX_CHAR_PATH],
    uint16_t processing_mode, 
    float radiometric_metadata[3], 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    struct extraterrestrial_solar_irradiance *esi, 
    struct local_angels *la);

int8_t radiometric_calibration_pixel(
    struct hsi_cube_pixel *cube_pixel, 
    struct calibrated_cube_pixel *calibrated_pixel, 
    uint16_t processing_mode, 
    float exposure, 
    struct Spectral_calibration *sc, 
    struct radiometric_calibration_coeff *rcc, 
    struct extraterrestrial_solar_irradiance *esi, 
    struct local_angels *la,
    float background_value, 
    float earth_sun_distance, 
    uint32_t radio_index, 
    uint16_t local_angels_index);


#endif

