/*
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
*/


#ifndef SVMBDT_H
#define SVMBDT_H

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#define COMPILE_DO_DEBUG 1
 
#if COMPILE_DO_DEBUG
#define DEBUG(str) str
#else
#define DEBUG(str)
#endif // COMPILE_DO_DEBUG


#define MAX_CLASSES 16
#define HYSPO_IMAGE_MAX_SIZE 956*684
#define SPECTRAL_BAND_SIZE 120
#define SVM_MODEL_DEFAULT_NAME_LENGTH 10
#define MAX_CHAR_PATH_LENGTH 200
#define MAX_FILENAME_LENGTH 80 
#define BDT_INDEX_LENGTH 3

#define LINEAR_SVM_MODEL_NAME "lsm"

#define SVM_MODELS_PATH "/svm_models/svm_models"
#define BDT_MODELS_PATH "/bdt/bdt"
#define SPARSE_BAND_PATH "/sparse_bands/sparse_bands"

#define LSM_RADIANCE "lsm_radiance"
#define LSM_RADIANCE_SPARSE "lsm_radiance_sparse"
#define LSM_REFLECTENCE "lsm_reflectence"
#define LSM_REFLECTENCE_SPARSE "lsm_reflectence_sparse"

#define RADIANCE "radiance"
#define RADIANCE_SPARSE "radiance_sparse"
#define REFLECTENCE "reflectence"
#define REFLECTENCE_SPARSE "reflectence_sparse"


struct Labeled_Image{ 
    uint32_t classification_execution_time; 
    uint32_t loading_execution_time; 
    uint8_t classes[MAX_CLASSES]; 
    uint8_t predicted_image[HYSPO_IMAGE_MAX_SIZE];  
};

int8_t svmbdt_init(
    char config_folder_path[MAX_CHAR_PATH_LENGTH], 
    uint8_t bdt_num, 
    uint8_t processing_mode, 
    uint8_t bdt_model[MAX_CLASSES][BDT_INDEX_LENGTH], 
    float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
    uint8_t svmbdt_metadata[1]);


int8_t svmbdt_predict(
    float X[SPECTRAL_BAND_SIZE], 
    uint8_t bdt[MAX_CLASSES][BDT_INDEX_LENGTH], 
    struct Labeled_Image* image, 
    float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
    uint8_t n_bands_used, 
    uint8_t processing_mode, 
    uint32_t index);


int8_t labeled_image_bin(char labels_save_path[MAX_CHAR_PATH_LENGTH],
    struct Labeled_Image *image,
    uint32_t n_pixels);


#endif


