/*
 * Author:    Jonas Gjendem Røysland
 * Created:   14.06.2023
*/


#include "../inc/svmbdt.h"

// Private functions

int8_t int_to_svm_model_name(uint16_t value, 
    char c1[3], 
    char c2[3]){

    int16_t cf1 = (int16_t)(value - value%16)/16; 
    int16_t cf2 = (int16_t)(value%16);
    char zero[2] = "0"; 
    char buffer[2]; 

    if(cf1 < 10){
        snprintf(c1,sizeof(&zero), zero); 
        snprintf(buffer, sizeof(buffer), "%i" , cf1); 
        strcat(c1,buffer);
    }

    else{
        snprintf(c1, sizeof(&c1), "%i", cf1);
    }

    if(cf2 < 10){
        snprintf(c2,sizeof(&zero), zero); 
        snprintf(buffer, sizeof(buffer), "%i" , cf2); 
        strcat(c2,buffer);
    }

    else{
        snprintf(c2, sizeof(&c2), "%i", cf2);
    }
     
    return 0; 
}


int8_t models_names(uint8_t bdt[MAX_CLASSES][BDT_INDEX_LENGTH],
        char model_name[SVM_MODEL_DEFAULT_NAME_LENGTH], 
        char svm_model_list[MAX_CLASSES][SVM_MODEL_DEFAULT_NAME_LENGTH]){

    char c1[3]; 
    char c2[3]; 
    char svm_model_name[20];

    for(uint16_t i = 0; i < MAX_CLASSES; i++){
        snprintf(svm_model_name, sizeof(&model_name), model_name); 

        int_to_svm_model_name(bdt[i][0],c1, c2);
        
        strcat(svm_model_name,c1);
        strcat(svm_model_name,c2);
        snprintf(svm_model_list[i], sizeof(svm_model_name), svm_model_name); 
    }

    return 0; 

} 


int8_t model_load(
    char svm_models_list[MAX_CLASSES][SVM_MODEL_DEFAULT_NAME_LENGTH],
    int16_t index,  
    float svm_models_array[MAX_CLASSES][SPECTRAL_BAND_SIZE+10],
    char model_path[MAX_CHAR_PATH_LENGTH]){

    float model_variables[SPECTRAL_BAND_SIZE+30];
    int16_t classes_1[2];
    int16_t classes_2[2]; 

    char filename[MAX_CHAR_PATH_LENGTH]; 
    
    FILE *ptr; 
    snprintf(filename, MAX_CHAR_PATH_LENGTH ,model_path); 
    strcat(filename,svm_models_list[index]);

    ptr = fopen(filename,"rb");

    if(ptr == NULL) {
        DEBUG(printf("SVMBDT: Loading svm_model: %s failed! \n", filename);)
        return -1; 
    }

    else{
        
        fread(&classes_1, 1, 1, ptr);
        svm_models_array[index][0] = classes_1[0]; 
 
        fread(&classes_2, 1, 1, ptr);
        svm_models_array[index][1] = classes_2[0]; 

        fread(&model_variables, SPECTRAL_BAND_SIZE+3, sizeof(float), ptr);

        for(uint16_t j = 2; j < SPECTRAL_BAND_SIZE+3; j++){
            svm_models_array[index][j] = model_variables[j-2]; 
        }

        fclose(ptr); 

        return 0; 
        
     }
}

int8_t generate_svm_model_name_folder(
    char config_folder_path[MAX_CHAR_PATH_LENGTH],
    uint16_t bdt_num, 
    uint16_t processing_mode, 
    char svm_models_path[MAX_CHAR_PATH_LENGTH]){

    char bdt_num_str[5];
    sprintf(bdt_num_str, "%d", bdt_num);

    strcpy(svm_models_path, config_folder_path);
    strcat(svm_models_path, SVM_MODELS_PATH);
    strcat(svm_models_path, bdt_num_str);
    strcat(svm_models_path, "/");

    if(processing_mode == 0){
        strcat(svm_models_path, LSM_RADIANCE);
    }
    
    if(processing_mode == 1){
        strcat(svm_models_path, LSM_RADIANCE_SPARSE);
    }
    
    if(processing_mode == 2){
        strcat(svm_models_path, LSM_REFLECTENCE);
    }
    
    if(processing_mode == 3){
        strcat(svm_models_path, LSM_REFLECTENCE_SPARSE);
    }

    strcat(svm_models_path, "/"); 

    return 0; 
    
}

int8_t generate_bdt_path(
    char config_folder_path[MAX_CHAR_PATH_LENGTH], 
    uint8_t bdt_num,
    char bdt_path[MAX_CHAR_PATH_LENGTH]){

    char bdt_num_char[3] = "";
    sprintf(bdt_num_char, "%d", bdt_num);

    strcpy(bdt_path, config_folder_path);
    strcat(bdt_path, BDT_MODELS_PATH);
    strcat(bdt_path, bdt_num_char);
        
    return 0; 
}


int8_t generate_sparse_band_path(
    char config_folder_path[MAX_CHAR_PATH_LENGTH],
    uint8_t bdt_num,
    char sparse_band_path[MAX_CHAR_PATH_LENGTH]
    ){

    char bdt_num_str[5] = "";
    sprintf(bdt_num_str, "%d", bdt_num);

    strcpy(sparse_band_path, config_folder_path); 
    strcat(sparse_band_path, SPARSE_BAND_PATH);
    strcat(sparse_band_path, bdt_num_str);

    return 0; 
}

int8_t labeled_image_name(char capture_path[MAX_CHAR_PATH_LENGTH],
    char labeled_image_save_folder[MAX_CHAR_PATH_LENGTH], 
    char labeled_image_file_path[MAX_CHAR_PATH_LENGTH],
    int16_t processing_mode){

    char capture_path_holder[MAX_CHAR_PATH_LENGTH];
    char name_holder[MAX_CHAR_PATH_LENGTH] = "";

    strcpy(labeled_image_file_path, labeled_image_save_folder); 
    strcpy(capture_path_holder, capture_path); 

    char * token = strtok(capture_path_holder, "/");
    while(token != NULL){
        if(token != NULL){
            strcpy(name_holder, token);
        }
        token = strtok(NULL, "/");
    }

    strcat(labeled_image_file_path, name_holder);

    strcat(labeled_image_file_path, "-labels-");

    if(processing_mode == 0){
        strcat(labeled_image_file_path, RADIANCE);
    }
    
    if(processing_mode == 1){
        strcat(labeled_image_file_path, RADIANCE_SPARSE);
    }
    
    if(processing_mode == 2){
        strcat(labeled_image_file_path, REFLECTENCE);
    }
    
    if(processing_mode == 3){
        strcat(labeled_image_file_path, REFLECTENCE_SPARSE);
    }

    return 0;
}



int8_t linear_svm_models_load(uint8_t bdt[MAX_CLASSES][BDT_INDEX_LENGTH],
                     float svm_models_array[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
                    char model_path[MAX_CHAR_PATH_LENGTH]){

    char svm_default_name[SVM_MODEL_DEFAULT_NAME_LENGTH]; 

    strncpy(svm_default_name,LINEAR_SVM_MODEL_NAME, SVM_MODEL_DEFAULT_NAME_LENGTH);

    char svm_models_list[MAX_CLASSES][SVM_MODEL_DEFAULT_NAME_LENGTH]; 
    models_names(bdt,svm_default_name, svm_models_list); 

    int8_t ret_t = 0; 
    uint8_t load_flag = 0; 

    for(uint16_t i = 0; i < MAX_CLASSES; i++){
        if(bdt[i][0] != 0){
            ret_t = model_load(svm_models_list, i, svm_models_array, model_path);
            if(ret_t < 0){
                load_flag = -1; 
            }
        }
    }

    if(load_flag < 0){
        return -1; 
    }

    else{
        return 0; 
    }

}

int8_t sparse_band_load(char path[MAX_CHAR_PATH_LENGTH], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE]){
    
    FILE *ptr; 
    ptr = fopen(path, "rb"); 

    if(ptr == NULL){
        DEBUG(printf("SVMBDT: Loading Sparse Band Failed! \n");)
        return -1;
    }

    else{

        for(uint16_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            fread(&spectral_band_list[i], 1, 1, ptr);
        }

        uint16_t n_bands_used = 0; 

        for(uint16_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
            if(spectral_band_list[i] != 0){
                n_bands_used++; 
            }
        }
    
        fclose(ptr);
        return n_bands_used; 
    }
}

int8_t load_bdt(char path[MAX_CHAR_PATH_LENGTH], 
    uint8_t bdt[MAX_CLASSES][BDT_INDEX_LENGTH]){
    
    FILE *ptr; 

    ptr = fopen(path, "rb");

    if(ptr == NULL) {
        DEBUG(printf("SVMBDT: Loading BDT Failed! \n");)
        return -1;
    }

    else{
        for(uint16_t i = 0; i < MAX_CLASSES; i++){
            for(uint16_t j = 0; j < BDT_INDEX_LENGTH; j++){
                bdt[i][j] = 0; 
            }
        }

        struct stat st;
        stat(path, &st);
        long sz = st.st_size;
    
        uint8_t n_classes = (uint8_t)(sz/BDT_INDEX_LENGTH);

        for(uint16_t i = 0; i < n_classes; i++){
            for(uint8_t j = 0; j < BDT_INDEX_LENGTH; j++){
                fread(&bdt[i][j], 1, 1, ptr);
            }
        }

        fclose(ptr); 

        return 0; 
    }
}


int8_t svm_linear(float X_i[SPECTRAL_BAND_SIZE], 
    float svm_model[SPECTRAL_BAND_SIZE+10]){

    float pred = 0.0; 

    for(uint16_t i = 0; i < SPECTRAL_BAND_SIZE; i++){
        pred += (X_i[i]*svm_model[3+i]) + svm_model[2]; 
    }

    if(pred >= 0){
        return -1; 
    }

    else{
        return 1; 
    }
}


int8_t sparse_svm(float X_i[SPECTRAL_BAND_SIZE], 
        float svm_model[SPECTRAL_BAND_SIZE+10], 
        uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
        uint8_t n_bands_used){

    float pred = 0.0; 

    for(uint16_t i = 0; i < n_bands_used; i++){

        pred += (X_i[spectral_band_list[i]]*svm_model[3+spectral_band_list[i]]) + svm_model[2]; 
    }

    if(pred >= 0){
        return -1; 
    }

    else{
        return 1; 
    }

}

int8_t classes_in_labels(struct Labeled_Image* image, uint32_t n_pixels){
    uint16_t index = 0; 

    for(uint16_t i = 0; i < MAX_CLASSES; i++){
        image->classes[i] = 255; //For differenciate no class. 
    }

    image->classes[0] = image->predicted_image[0];

    for(uint32_t i = 1; i < n_pixels; i++){
        uint16_t check = 0; 
        if(image->classes[index] != image->predicted_image[i]){

            for(uint16_t j = 0; j < MAX_CLASSES; j++){
                if(image->classes[j] == image->predicted_image[i]){
                    check = 1; 
                }
            }

            if(check == 0){
                image->classes[index+1] = image->predicted_image[i];
                index++;
            }      
        }
    }

    //Sorting classes array

    uint8_t class_holder; 

    for(uint16_t i = 0; i < MAX_CLASSES-1; i++){
        if(image->classes[i] > image->classes[i+1]){
            class_holder = image->classes[i];
            image->classes[i] = image->classes[i+1];
            image->classes[i+1] = class_holder; 
        }
    }

    return 0; 
}



int8_t n_classes(struct Labeled_Image *image){
    uint8_t counter = 0; 

    for(uint16_t i = 0; i < MAX_CLASSES; i++){
        if(image->classes[i] == 255){
            break; 
        }

        else{
            counter++; 
        }
    }

    return counter; 

}


int8_t labeles_to_compressed_format(struct Labeled_Image *image, uint16_t n_class){
    
    for(uint32_t i = 0; i < HYSPO_IMAGE_MAX_SIZE; i++){
        for(uint32_t j = 0; j < n_class; j++){
            if(image->predicted_image[i] == image->classes[j]){
                image->predicted_image[i] = j;
                break; 
            }
        }
    }

    return 0; 
}


//Public function

int8_t labeled_image_bin(char labels_save_path[MAX_CHAR_PATH_LENGTH],
    struct Labeled_Image *image,
    uint32_t n_pixels){
    
    FILE *ptr; 

    ptr = fopen(labels_save_path, "wb");
    unsigned char buffer[1]; 

    if(ptr == NULL) {
        DEBUG(printf("File creating: something went wrong \n");)
        return -1;
    }

    classes_in_labels(image,n_pixels); 

    uint8_t n_class = n_classes(image); 
    labeles_to_compressed_format(image, n_class);

    fwrite(&image->classification_execution_time,sizeof(uint32_t),1,ptr);
    fwrite(&image->loading_execution_time,sizeof(uint32_t),1,ptr);
    fwrite(image->classes, sizeof(uint8_t),16,ptr); 

    if(n_class <= 2){
        uint8_t pixels_1; 

        for(uint32_t i = 0; i < n_pixels; i = i + 8){

            pixels_1 &= 0x00;

            pixels_1 |= (image->predicted_image[i] << 7); 
            pixels_1 |= (image->predicted_image[i+1] << 6); 
            pixels_1 |= (image->predicted_image[i+2] << 5); 
            pixels_1 |= (image->predicted_image[i+3] << 4); 
            pixels_1 |= (image->predicted_image[i+4] << 3); 
            pixels_1 |= (image->predicted_image[i+5] << 2); 
            pixels_1 |= (image->predicted_image[i+6] << 1); 
            pixels_1 |= (image->predicted_image[i+7] << 0); 

            fwrite(&pixels_1,1,1,ptr);
        }
    }

    if(n_class > 2 && n_class <= 4){
        //struct four_classes pixel_byte;
        uint8_t pixels_2; 


        for(uint32_t i = 0; i < n_pixels; i = i + 4){

            pixels_2 &= 0x00;

            pixels_2 |= (image->predicted_image[i] << 6); 
            pixels_2 |= (image->predicted_image[i+1] << 4); 
            pixels_2 |= (image->predicted_image[i+2] << 2); 
            pixels_2 |= (image->predicted_image[i+3] << 0); 

            fwrite(&pixels_2,1,1,ptr);

        }
    }

    if(n_class > 4 && n_class <= MAX_CLASSES){
        uint8_t pixels_4;

        for(uint32_t i = 0; i < n_pixels; i = i + 2){

            pixels_4 &= 0x00;

            pixels_4 |= (image->predicted_image[i] << 4); 
            pixels_4 |= (image->predicted_image[i+1] << 0); 

            fwrite(&pixels_4,1,1,ptr);
        }
    }

    fclose(ptr);

    return 0; 

}


int8_t svmbdt_init(
    char config_folder_path[MAX_CHAR_PATH_LENGTH], 
    uint8_t bdt_num, 
    uint8_t processing_mode, 
    uint8_t bdt_model[MAX_CLASSES][BDT_INDEX_LENGTH], 
    float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
    uint8_t svmbdt_metadata[1]){

    int8_t ret_bdt = 0;
    int8_t ret_model = 0; 
    int8_t ret_sparse = 0;  

    char bdt_path[MAX_CHAR_PATH_LENGTH] = ""; 
    char svm_models_path[MAX_CHAR_PATH_LENGTH] = ""; 

    generate_bdt_path(config_folder_path, bdt_num, bdt_path); 
    generate_svm_model_name_folder(config_folder_path, bdt_num, processing_mode, svm_models_path);

    ret_bdt = load_bdt(bdt_path, bdt_model);
    ret_model = linear_svm_models_load(bdt_model, svm_models, svm_models_path);
    
    if(processing_mode == 1 || processing_mode == 3){
        char sparse_band_path[MAX_CHAR_PATH_LENGTH] = ""; 

        ret_sparse = generate_sparse_band_path(config_folder_path, bdt_num, sparse_band_path);

        svmbdt_metadata[0] = sparse_band_load(sparse_band_path, spectral_band_list);

    }

    if(ret_bdt < 0 || ret_model < 0 || ret_sparse < 0){
        DEBUG(printf("SVMBDT: Something went wrong loading config data! \n");)
        return -1; 
    }

    else{
        return 0; 
    }

}

int8_t svmbdt_predict(
    float X[SPECTRAL_BAND_SIZE], 
    uint8_t bdt[MAX_CLASSES][BDT_INDEX_LENGTH], 
    struct Labeled_Image* image, 
    float svm_models[MAX_CLASSES][SPECTRAL_BAND_SIZE+10], 
    uint8_t spectral_band_list[SPECTRAL_BAND_SIZE], 
    uint8_t n_bands_used, 
    uint8_t processing_mode, 
    uint32_t index){

    uint16_t svm_models_index = 0; 
    uint16_t previous_index = 0; 
    int8_t pixel_holder = 0; 
    uint8_t safe_counter = 0; 

    if(processing_mode == 1 || processing_mode == 3){

        //Sparse SVM
    
            while(1){

                previous_index = svm_models_index; 

                pixel_holder = sparse_svm(X, svm_models[svm_models_index], 
                                spectral_band_list, n_bands_used); 

                if(pixel_holder >= 0){
                    svm_models_index = bdt[svm_models_index][1]; 
                }
                else{
                    svm_models_index = bdt[svm_models_index][2]; 
                }

                safe_counter++; 

                if(safe_counter > MAX_CLASSES){
                    break; 
                }

                if(svm_models_index == 0){

                    if(pixel_holder >= 0){
                        image->predicted_image[index] = svm_models[previous_index][0];
                    }

                    else{
                        image->predicted_image[index] = svm_models[previous_index][1]; 
                    }

                    break; 
                }
            }
    }

    else{

        //Linear SVM

            while(1){

                previous_index = svm_models_index; 

                pixel_holder = svm_linear(X, svm_models[svm_models_index]); 

                if(pixel_holder >= 0){
                    svm_models_index = bdt[svm_models_index][1]; 
                }
                else{
                    svm_models_index = bdt[svm_models_index][2]; 
                }

                safe_counter++; 

                if(safe_counter > MAX_CLASSES){
                    break; 
                }

                if(svm_models_index == 0){

                    if(pixel_holder >= 0){
                        image->predicted_image[index] = svm_models[previous_index][0];
                    }

                    else{
                        image->predicted_image[index] = svm_models[previous_index][1]; 
                    }

                    break; 
                }
            }
    }
    
    return 0; 

}






 

/*
0 radiance
1 radiance_sparse 
3 reflectence 
4 reflectence_sparse 
*/
